# -*- coding: UTF-8 -*-
from datetime import datetime
from operator import itemgetter

import pickle
import glob
import os
import sys
import time
import random
import pytest
import re
import warnings

if sys.version_info < (3, 0):
    modes = ('r', 'w', )
else:
    modes = ('rb', 'wb',)


# Reference: http://docs.gurock.com/testrail-api2/reference-statuses
TESTRAIL_TEST_STATUS = {
    "passed": 1,
    "blocked": 2,
    "untested": 3,
    "retest": 4,
    "failed": 9
}

TESTRAIL_TEST_STATUS_IDS = {v: k for k, v in TESTRAIL_TEST_STATUS.items()}

PYTEST_TO_TESTRAIL_STATUS = {
    "passed": TESTRAIL_TEST_STATUS["passed"],
    "failed": TESTRAIL_TEST_STATUS["failed"],
    "skipped": TESTRAIL_TEST_STATUS["blocked"],
}

DT_FORMAT = '%d-%m-%Y %H:%M:%S'

TESTRAIL_PREFIX = 'testrail'
PYTEST_PARAMETRIZE = 'parametrize'

ADD_RESULT_URL = 'add_result_for_case/{}/{}'
ADD_TESTRUN_URL = 'add_run/{}'
CLOSE_TESTRUN_URL = 'close_run/{}'
CLOSE_TESTPLAN_URL = 'close_plan/{}'
GET_TESTRUN_URL = 'get_run/{}'
GET_TESTPLAN_URL = 'get_plan/{}'
GET_TESTS_URL = 'get_tests/{}'
GET_TEST_CASE_URL = 'get_case/{}'
UPDATE_TEST_CASE_URL = 'update_case/{}'

COMMENT_SIZE_LIMIT = 4000


class DeprecatedTestDecorator(DeprecationWarning):
    pass


warnings.simplefilter(action='once', category=DeprecatedTestDecorator, lineno=0)


class pytestrail(object):
    '''
    An alternative to using the testrail function as a decorator for test cases, since py.test may confuse it as a test
    function since it has the 'test' prefix
    '''

    @staticmethod
    def case(*ids, **kwargs):
        """
        Decorator to mark tests with testcase ids.

        ie. @pytestrail.case('C123', 'C12345')

        :return pytest.mark:
        """
        return pytest.mark.testrail(ids=ids)


def testrail(*ids):
    """
    Decorator to mark tests with testcase ids.

    ie. @testrail('C123', 'C12345')

    :return pytest.mark:
    """
    deprecation_msg = ('pytest_testrail: the @testrail decorator is deprecated and will be removed. Please use the '
                       '@pytestrail.case decorator instead.')
    warnings.warn(deprecation_msg, DeprecatedTestDecorator)
    return pytestrail.case(*ids)


def get_test_outcome(outcome):
    """
    Return numerical value of test outcome.

    :param str outcome: pytest reported test outcome value.
    :returns: int relating to test outcome.
    """
    return PYTEST_TO_TESTRAIL_STATUS[outcome]


def testrun_name():
    """Returns testrun name with timestamp"""
    now = datetime.today()
    return 'Automated Run {}'.format(now.strftime(DT_FORMAT))


def clean_test_ids(test_ids):
    """
    Clean pytest marker containing testrail testcase ids.

    :param list test_ids: list of test_ids.
    :return list ints: contains list of test_ids as ints.
    """
    def _format_case_id(row):
        try:
            return int(re.search('(?P<test_id>[0-9]+$)', row).groupdict().get('test_id'))
        except Exception as e:
            return None

    return [_format_case_id(test_id) for test_id in test_ids if _format_case_id(test_id)]


def get_marker(item, marker):
    py_version = [int(i) for i in pytest.__version__.split('.')]
    if py_version < [4, 0]:
        markers = item.get_marker(marker)
    else:
        markers = None
        from _pytest.mark.structures import Mark

        class MarkMutable(Mark):
            _marks = []

        _markers = item.get_closest_marker(marker)
        if _markers:
            markers = MarkMutable(**{k: v for k, v in _markers.__dict__.items() if not k.startswith("_")})
            markers._marks.extend([i for i in item.iter_markers(marker)])
    return markers


def get_testrail_keys(items):
    """Return Tuple of Pytest nodes and TestRail ids from pytests markers"""
    def __get_tr_into_parametrize(__params):
        __test_ids = []
        if isinstance(__params, dict):
            for k, v in __params.items():
                if k.lower() == 'testrailcase':
                    if isinstance(v, str):
                        __test_ids.append(v)
                    elif isinstance(v, list) or isinstance(v, tuple):
                        __test_ids.extend(v)
                else:
                    __test_ids.extend(__get_tr_into_parametrize(v))
        elif isinstance(__params, list) or isinstance(__params, tuple):
            for index, row in enumerate(__params):
                __test_ids.extend(__get_tr_into_parametrize(row))
        else:
            pass
        return list(set(__test_ids))

    testcaseids = []
    for item in items:
        __ids = []
        if get_marker(item, TESTRAIL_PREFIX):
            __ids.extend(get_marker(item, TESTRAIL_PREFIX).kwargs.get('ids'))
        if get_marker(item, PYTEST_PARAMETRIZE):
            bundle_params = item.callspec.params
            all_params = get_marker(item, PYTEST_PARAMETRIZE)._marks
            for param_name, param_value in bundle_params.items():
                if isinstance(param_value, dict):
                    __ids.extend(__get_tr_into_parametrize(param_value))
                else:
                    for param in all_params:
                        if param_name == param.args[0] and isinstance(param.args[1], dict):
                            __ids.extend(__get_tr_into_parametrize(param.args[1][param_value]))
        if __ids:
            testcaseids.append((item, clean_test_ids(list(set(__ids)))))
    return testcaseids


class PyTestRailPlugin(object):
    def __init__(self, client, assign_user_id, project_id, suite_id, include_all, cert_check, tr_name, bundle, is_bundle_end, run_id=0,
                 plan_id=0, version='', close_on_complete=False, publish_blocked=True, skip_missing=False):
        self.assign_user_id = assign_user_id
        self.cert_check = cert_check
        self.client = client
        self.project_id = project_id
        self.results = []
        self.tr_keys = []
        self.suite_id = suite_id
        self.include_all = include_all
        self.testrun_name = tr_name
        self.testrun_id = run_id
        self.testplan_id = plan_id
        self.version = version
        self.close_on_complete = close_on_complete
        self.publish_blocked = publish_blocked
        self.skip_missing = skip_missing
        self.bundle = bundle
        self.is_bundle_end = is_bundle_end
        self.node_id = None
        self.node_obj = dict(status='new', tr_keys=[], results=self.results, bundle=self.bundle)

    # pytest hooks

    def pytest_report_header(self, config, startdir):
        """ Add extra-info in header """
        message = 'pytest-testrail: '
        if self.testplan_id:
            message += 'existing testplan #{} selected'.format(self.testplan_id)
        elif self.testrun_id:
            message += 'existing testrun #{} selected'.format(self.testrun_id)
        else:
            message += 'a new testrun will be created'
        message += ', nodeId #{} created'.format(self.node_id)
        return message

    def update_test_cases(self, tr_keys):
        def __analyze_test_case_info(info):
            __need_update = {}
            if info['custom_automation_type'] != 1:
                __need_update['custom_automation_type'] = 1
            if 1 not in info.get('custom_custom_labels', []):
                __labels = info.get('custom_custom_labels', [])
                __labels.append(1)
                __need_update['custom_custom_labels'] = list(set(__labels))
            if info['type_id'] != 3:
                __need_update['type_id'] = 3
            return __need_update

        for tr in tr_keys:
            response = self.client.send_get(
                GET_TEST_CASE_URL.format(tr),
                cert_check=self.cert_check
            )
            error = self.client.get_error(response)
            if error:
                print("[{}] Info: Can't get info by Testcase #{}, reason: '{}'".format(TESTRAIL_PREFIX, tr, error))
            else:
                data = __analyze_test_case_info(response)
                if data:
                    r = self.client.send_post(
                        UPDATE_TEST_CASE_URL.format(tr),
                        data,
                        cert_check=self.cert_check
                    )
                    err = self.client.get_error(r)
                    if err:
                        print("[{}] Info: Can't update info by Testcase #{}, reason: '{}'".format(TESTRAIL_PREFIX, tr, error))

    @pytest.hookimpl(trylast=True)
    def pytest_collection_modifyitems(self, session, config, items):
        self.create_node()
        items_with_tr_keys = get_testrail_keys(items)
        tr_keys = list(set([case_id for item in items_with_tr_keys for case_id in item[1]]))
        self.tr_keys = tr_keys
        if self.testplan_id and self.is_testplan_available():
            self.testrun_id = 0
        elif self.testrun_id and self.is_testrun_available():
            self.testplan_id = 0
            if self.skip_missing:
                tests_list = [
                    test.get('case_id') for test in self.get_tests(self.testrun_id)
                ]
                for item, case_id in items_with_tr_keys:
                    if not set(case_id).intersection(set(tests_list)):
                        mark = pytest.mark.skip('Test is not present in testrun.')
                        item.add_marker(mark)
        else:
            if self.testrun_name is None:
                self.testrun_name = testrun_name()
            else:
                now = datetime.today()
                self.testrun_name = '{} {}'.format(self.testrun_name, now.strftime(DT_FORMAT))

        self.save_node_data(self.node_id, tr_keys=tr_keys, status='testing')

    @pytest.hookimpl(tryfirst=True, hookwrapper=True)
    def pytest_runtest_makereport(self, item, call):
        """ Collect result and associated testcases (TestRail) of an execution """
        outcome = yield
        rep = outcome.get_result()
        comment=dict(result=None, custom=None)

        if get_marker(item, TESTRAIL_PREFIX) or get_testrail_keys([item]):
            print("When: {}".format(rep.when))
            testcaseids = get_testrail_keys([item])
            tr_keys = [case_id for _item in testcaseids for case_id in _item[1]]
            if get_marker(item, 'comment'):
                comment['custom'] = get_marker(item, 'comment').kwargs.get('comment', 'Empty comment')
            else:
                comment['custom'] = 'Not set'

            if rep.longrepr:
                if rep.capstdout:
                    comment['result'] = rep.capstdout
                else:
                    comment['result'] = rep.caplog

            if get_marker(item, PYTEST_PARAMETRIZE):
                print("Test is parametrized: {}".format(get_marker(item, PYTEST_PARAMETRIZE)))
                comment['custom'] += "\n This test is parametrized"

            if rep.when == 'call' and testcaseids:
                if not rep.outcome == 'skipped':
                    self.add_result(
                        tr_keys,
                        get_test_outcome(rep.outcome),
                        comment=comment,
                        duration=rep.duration
                    )
                else:
                    pass

    def pytest_sessionfinish(self, session, exitstatus):
        """ Publish results in TestRail """
        print('[{}] Start publishing'.format(TESTRAIL_PREFIX))
        if self.node_id:
            self.save_node_data(self.node_id, results=self.results, status='finish')
            if self.is_bundle_end:
                is_master = self.check_master_node(self.node_id)
                if self.results and is_master is True:
                    #print(self.results)
                    tests_list = [str(result['case_id']) for result in self.results]
                    print('[{}] Testcases to publish: {}'.format(TESTRAIL_PREFIX, ', '.join(tests_list)))
                    if self.testrun_id:
                        self.add_results(self.testrun_id)
                    elif self.testplan_id:
                        testruns = self.get_available_testruns(self.testplan_id)
                        print('[{}] Testruns to update: {}'.format(TESTRAIL_PREFIX, ', '.join([str(elt) for elt in testruns])))
                        for testrun_id in testruns:
                            self.add_results(testrun_id)
                    else:
                        print('[{}] No data published'.format(TESTRAIL_PREFIX))

                    if self.close_on_complete and self.testrun_id:
                        self.close_test_run(self.testrun_id)
                    elif self.close_on_complete and self.testplan_id:
                        self.close_test_plan(self.testplan_id)
                    self.clear_node_files()
        print('[{}] End publishing'.format(TESTRAIL_PREFIX))
    # plugin

    def add_result(self, test_ids, status, comment, duration=0):
        """
        Add a new result to results dict to be submitted at the end.

        :param list test_ids: list of test_ids.
        :param int status: status code of test (pass or fail).
        :param comment: None or a failure representation.
        :param duration: Time it took to run just the test.
        """
        comment['custom'] += ' [%s]' % TESTRAIL_TEST_STATUS_IDS[status]

        for test_id in test_ids:
            data = {
                'case_id': test_id,
                'status_id': status,
                'comment': comment,
                'duration': duration
            }
            #print("Add test-result: {}".format(data))
            self.results.append(data)

    def add_results(self, testrun_id):
        """
        Add results one by one to improve errors handling.

        :param testrun_id: Id of the testrun to feed

        """
        # unicode converter for compatibility of python 2 and 3
        try:
            converter = unicode
        except NameError:
            converter = lambda s, c: str(bytes(s, "utf-8"), c)
        # Results are sorted by 'case_id' and by 'status_id' (worst result at the end)
        self.results.sort(key=itemgetter('status_id'))
        self.results.sort(key=itemgetter('case_id'))

        # Manage case of "blocked" testcases
        if self.publish_blocked is False:
            print('[{}] Option "Don\'t publish blocked testcases" activated'.format(TESTRAIL_PREFIX))
            blocked_tests_list = [
                test.get('case_id') for test in self.get_tests(testrun_id)
                if test.get('status_id') == TESTRAIL_TEST_STATUS["blocked"]
            ]
            print('[{}] Blocked testcases excluded: {}'.format(TESTRAIL_PREFIX, ', '.join(str(elt) for elt in blocked_tests_list)))
            self.results = [result for result in self.results if result.get('case_id') not in blocked_tests_list]

        # prompt enabling include all test cases from test suite when creating test run
        if self.include_all:
            print('[{}] Option "Include all testcases from test suite for test run" activated'.format(TESTRAIL_PREFIX))

        # Publish results
        for i, result in enumerate(sorted(self.results, key=lambda result: result["status_id"])):
            print("Publish test-result: {}".format(result))
            comment = result['comment']['custom']
            test_result = result['comment']['result']
            data = {'status_id': result['status_id']}
            if self.version:
                data['version'] = self.version
            if comment:
                # Indent text to avoid string formatting by TestRail. Limit size of comment.
                data['custom_comment_custom'] = u"# Pytest result: #\n"
                data['custom_comment_custom'] += u'Log truncated\n...\n' if len(
                    converter(comment, 'utf8')) > COMMENT_SIZE_LIMIT else u''
                data['custom_comment_custom'] += u"    " + converter(comment, 'utf8')[
                                                           -COMMENT_SIZE_LIMIT:].replace('\n', '\n    ')
            if test_result:
                data['comment'] = test_result
            duration = result.get('duration')
            if duration:
                duration = 1 if (duration < 1) else int(round(duration))  # TestRail API doesn't manage milliseconds
                data['elapsed'] = str(duration) + 's'
            #print("Publish test-result (data to request): {}".format(data))
            response = self.client.send_post(
                ADD_RESULT_URL.format(testrun_id, result['case_id']),
                data,
                cert_check=self.cert_check
            )
            error = self.client.get_error(response)
            if error:
                print('[{}] Info: Testcase #{} not published for following reason: "{}"'.format(TESTRAIL_PREFIX, result['case_id'], error))

    def create_test_run(
            self, assign_user_id, project_id, suite_id, include_all, testrun_name, tr_keys):
        """
        Create testrun with ids collected from markers.

        :param tr_keys: collected testrail ids.
        """
        data = {
            'suite_id': suite_id,
            'name': testrun_name,
            'assignedto_id': assign_user_id,
            'include_all': include_all,
            'case_ids': tr_keys,
        }

        response = self.client.send_post(
            ADD_TESTRUN_URL.format(project_id),
            data,
            cert_check=self.cert_check
        )
        error = self.client.get_error(response)
        if error:
            print('[{}] Failed to create testrun: "{}"'.format(TESTRAIL_PREFIX, error))
        else:
            self.testrun_id = response['id']
            print('[{}] New testrun created with name "{}" and ID={}'.format(TESTRAIL_PREFIX, testrun_name, self.testrun_id))

    def create_node(self):
        time.sleep(random.randrange(100, 1000, 1) / 1000)
        self.node_id = format(float(time.time()), 'f')
        self.save_node_file(self.node_id, self.node_obj)

    def save_node_data(self, node_id, **kwargs):
        data = self.open_node_file(node_id)
        for k, v in kwargs.items():
            data[k] = v
            self.node_obj[k] = v
        self.save_node_file(node_id, data)

    def open_node_file(self, node_id):
        f = os.path.abspath('{}.node'.format(node_id))
        with open(f, modes[0]) as node_file:
            _data = node_file.read()
        if _data:
            return pickle.loads(_data)
        else:
            return None

    def save_node_file(self, node_id, data):
        f = os.path.abspath('{}.node'.format(node_id))
        _data = pickle.dumps(data)
        with open(f, modes[1]) as node_file:
            node_file.write(_data)

    def check_master_node(self, node_id):
        nodes = []
        for file in glob.glob("*.node"):
            __node_id = file.replace('.node', '')
            data = self.open_node_file(__node_id)
            if self.bundle == data['bundle']:
                nodes.append(__node_id)
        print("Current node_id {}. List node_ids: {}".format(node_id, nodes))
        if not list(filter(lambda x: float(node_id) < float(x), nodes)):
            print("I'm master node!")
            all_nodes_data = self.wait_all_nodes()
            self.merge_node_results(all_nodes_data)
            self.update_test_cases(self.tr_keys)
            return True
        else:
            print("I'm slave node. Waiting master execution...")
            return False

    def wait_all_nodes(self):
        while True:
            all_nodes_data = []
            for file in glob.glob("*.node"):
                node_id = file.replace('.node', '')
                data = self.open_node_file(node_id)
                if self.node_id != node_id and self.bundle == data['bundle']:
                    all_nodes_data.append(data)
            if len(all_nodes_data) == len(list(filter(lambda x: x['status'] == 'finish', all_nodes_data))):
                print("All nodes are 'finish' status!")
                return all_nodes_data
            else:
                print("Waiting all nodes 'finish' status")
                time.sleep(1)

    def merge_node_results(self, all_nodes_data):
        nodes_results = []
        for node_result in all_nodes_data:
            nodes_results.extend(node_result['results'])
        self.results.extend(nodes_results)

        nodes_tr_keys = []
        for result in self.results:
            case_id = result['case_id']
            if isinstance(case_id, list):
                nodes_tr_keys.extend(case_id)
            else:
                nodes_tr_keys.append(case_id)
        #print("Results collection: {}".format(self.results))
        if not self.testrun_id:
            self.create_test_run(
                self.assign_user_id,
                self.project_id,
                self.suite_id,
                self.include_all,
                self.testrun_name,
                nodes_tr_keys
            )

    def clear_node_files(self):
        for file in glob.glob("*.node"):
            os.remove(os.path.abspath(file))
            print("Remove node file {}".format(file))

    def close_test_run(self, testrun_id):
        """
        Closes testrun.

        """
        response = self.client.send_post(
            CLOSE_TESTRUN_URL.format(testrun_id),
            data={},
            cert_check=self.cert_check
        )
        error = self.client.get_error(response)
        if error:
            print('[{}] Failed to close test run: "{}"'.format(TESTRAIL_PREFIX, error))
        else:
            print('[{}] Test run with ID={} was closed'.format(TESTRAIL_PREFIX, self.testrun_id))

    def close_test_plan(self, testplan_id):
        """
        Closes testrun.

        """
        response = self.client.send_post(
            CLOSE_TESTPLAN_URL.format(testplan_id),
            data={},
            cert_check=self.cert_check
        )
        error = self.client.get_error(response)
        if error:
            print('[{}] Failed to close test plan: "{}"'.format(TESTRAIL_PREFIX, error))
        else:
            print('[{}] Test plan with ID={} was closed'.format(TESTRAIL_PREFIX, self.testplan_id))

    def is_testrun_available(self):
        """
        Ask if testrun is available in TestRail.

        :return: True if testrun exists AND is open
        """
        response = self.client.send_get(
            GET_TESTRUN_URL.format(self.testrun_id),
            cert_check=self.cert_check
        )
        error = self.client.get_error(response)
        if error:
            print('[{}] Failed to retrieve testrun: "{}"'.format(TESTRAIL_PREFIX, error))
            return False

        return response['is_completed'] is False

    def is_testplan_available(self):
        """
        Ask if testplan is available in TestRail.

        :return: True if testplan exists AND is open
        """
        response = self.client.send_get(
            GET_TESTPLAN_URL.format(self.testplan_id),
            cert_check=self.cert_check
        )
        error = self.client.get_error(response)
        if error:
            print('[{}] Failed to retrieve testplan: "{}"'.format(TESTRAIL_PREFIX, error))
            return False

        return response['is_completed'] is False

    def get_available_testruns(self, plan_id):
        """
        :return: a list of available testruns associated to a testplan in TestRail.

        """
        testruns_list = []
        response = self.client.send_get(
            GET_TESTPLAN_URL.format(plan_id),
            cert_check=self.cert_check
        )
        error = self.client.get_error(response)
        if error:
            print('[{}] Failed to retrieve testplan: "{}"'.format(TESTRAIL_PREFIX, error))
        else:
            for entry in response['entries']:
                for run in entry['runs']:
                    if not run['is_completed']:
                        testruns_list.append(run['id'])
        return testruns_list

    def get_tests(self, run_id):
        """
        :return: the list of tests containing in a testrun.

        """
        response = self.client.send_get(
            GET_TESTS_URL.format(run_id),
            cert_check=self.cert_check
        )
        error = self.client.get_error(response)
        if error:
            print('[{}] Failed to get tests: "{}"'.format(TESTRAIL_PREFIX, error))
            return None
        return response
